unit IC_LMM;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	IC Location Manager Module
	--------------------------
	This file implements a Location Manager Module (LMM) for switching between
	IC profiles.  The implementation follows fairly standard LMM lines.
*)

interface

	uses
		Components;

{$main}
	function Main(var params : ComponentParameters; storage : Ptr) : ComponentResult;

implementation

	uses
		Resources,
		LocationManager,
		Script,
		TextUtils,
		Fonts,
		LowMem,

		PascalA4,

		InternetConfig,
		
		ICDebug,
		ICCommonSubs,
		ICStringLists;

	(* ***** Broken Apple Interfaces ***** *)

	// Location Manager selectors that should be defined in
	// "LocationManager.p" but aren't because, in the C interfaces, they
	// are defined in "LocationManager.k.h" (WTF does the ".k" mean), which
	// is not ported to Pascal, so the constants never made it to Pascal
	// either.  I've already filed this as a bug.
	
	const
		kALMGetCurrentSelect = $0000;
		kALMSetCurrentSelect = $0001;
		kALMCompareSettingSelect = $0002;
		kALMEditSettingSelect = $0003;
		kALMDescribeSettingsSelect = $0004;
		kALMDescribeErrorSelect = $0005;
		kALMImportExportSelect = $0006;
		kALMGetScriptInfoSelect = $0007;
		kALMGetInfoSelect = $0008;
		
		kALMFirstSelect = kALMGetCurrentSelect;
		kALMLastSelect = kALMGetInfoSelect;

	(* ***** Resources ***** *)
	
	const
		// 'TEXT'
		
		rInfoText = 128;
		
		// 'STR#'
		
		rMiscStrings = 128;
			strInternetConfigSet = 1;
			strNoICErr  = 2;
			strInsufficientICErr = 3;
			strNoMemoryErr = 4;
			strCantFindICErr = 5;
			strCantHackIt = 6;
			strProfileNotFoundErr = 7;
			strInvalidSetting = 8;
			strSettingTooOld = 9;
			strSettingTooNew = 10;
			strConfirmNamePrompt = 11;
			
		// 'SINF'
		
		rOurScriptInfo = 128;
			
	(* ***** Misc Constants ***** *)

	const
		kICLMMCreator = 'IClM';
	
	// Location Manager suggests that you return positive fake error codes
	// when you hit a specific problem and you wish to ALM to call your
	// ALMDescribeError routine to get the correct error text.  The
	// following as the ones used by IC LMM.
	
	const
		memFullFakeErr = 1;
		afpItemNotFoundFakeErr = 2;
		noPortFakeErr = 3;
		invalidSettingFakeErr = 4;
		settingTooOldFakeErr = 5;
		settingTooNewFakeErr = 6;
		noICFakeErr = 7;
		insufficientICFakeErr = 8;
			
	(* ***** Globals ***** *)

	(*	Our globals are very simple.  The only data we maintain is
		a connection to IC and a connection to our resources. These
		are manipulated by the various global utility routines which
		follow.
		
		Note that in both cases we use a 'lazy' open technique.
		The component only opens a connection to IC when it actually
		needs data from it.  Similarly with the components resources.
		This helps make our ICLMMOpen routine faster, especially given
		that ALM is prone to opening an instance of us, calling a single
		entry point, and then closing it again.
	*)
	
	type
		ComponentGlobals = 
			record
				self : ComponentInstance;
				target : ComponentInstance;
				inst : ICInstance;
				ourResFile : integer;
			end;
		ComponentGlobalsPtr = ^ComponentGlobals;
	
	(* ***** Global Utilities ***** *)
	
	function StartUsingIC(globals : ComponentGlobalsPtr) : OSStatus;
		(*	This routine is called when the component needs to
			access Internet Config.  We perform a 'lazy' open
			of IC, as described above.  The instance is closed
			when we're closed, ie in ICLMMClose.
		*)
		var
			err : OSStatus;
			icVersion : longint;
	begin
		err := noErr;
		if globals^.inst = nil then begin
			err := ICStart(globals^.inst, kICLMMCreator);
			if err = badComponentInstance then begin
				err := noICFakeErr;
			end; (* if *)
			if err = noErr then begin
				icVersion := GetComponentVersion(ComponentInstance(globals^.inst));
				if band(icVersion, $FFFF0000) < kICComponentInterfaceVersion3 then begin
					err := insufficientICFakeErr;
				end; (* if *)
			end; (* if *)
			if err = noErr then begin
				err := ICFindConfigFile(globals^.inst, 0, nil);
			end; (* if *)
		end; (* if *)
		StartUsingIC := err;
	end; (* StartUsingIC *)
	
	function BeginUsingResources(globals : ComponentGlobalsPtr; var oldResFile : integer) : OSStatus;
		(*	This routine is called when the component needs to
			access its own resources.  We perform a 'lazy' open
			of our resource fork, as described above.  Unlike our
			connection to IC, the client must be sure to call
			EndUsingResources (although ICLMMClose does contain
			an safeguard).  This is because I didn't want to leave
			the resource file open, even though ALM says its OK.
		*)
		var
			err : OSStatus;
	begin
		err := noErr;
		oldResFile := CurResFile;
		if globals^.ourResFile = 0 then begin
			globals^.ourResFile := OpenComponentResFile(Component(globals^.self));
			err := ResError;
		end else begin
			UseResFile(globals^.ourResFile);
			ICAssert(ResError = noErr);
		end; (* if *)
		BeginUsingResources := err;
	end; (* BeginUsingResources *)
	
	procedure EndUsingResources(oldResFile : integer);
		(*	Finish using our resources. *)
	begin
		if oldResFile <> 0 then begin
			UseResFile(oldResFile);
			ICAssert(ResError = noErr);
		end; (* if *)
	end; (* EndUsingResources *)

	(* ***** Component Manager Routines ***** *)
	
	(* Routines that are generally implemented by components. *)

	function ICLMMCanDo(globals : ComponentGlobalsPtr; selector : integer) : ComponentResult;
	begin
		{$unused globals}
		case selector of
			kComponentVersionSelect,
				kComponentCanDoSelect,
					kComponentOpenSelect,
					kComponentCloseSelect,
					kComponentTargetSelect,
					kALMFirstSelect..kALMLastSelect:
				ICLMMCanDo := 1;
			otherwise
				ICLMMCanDo := 0;
		end; (* case *)
	end; (* ICLMMCanDo *)

	function ICLMMClose(globals : ComponentGlobalsPtr; self : ComponentInstance) : ComponentResult;
		var
			err : ComponentResult;
			junk : OSStatus;
	begin
		{$unused self}
		err := noErr;
		if globals <> nil then begin
			if globals^.inst <> nil then begin
				junk := ICStop(globals^.inst);
				globals^.inst := nil;
				ICAssert(junk = noErr);
			end; (* if *)
			if globals^.ourResFile <> 0 then begin
				junk := CloseComponentResFile(globals^.ourResFile);
				globals^.ourResFile := 0;
				ICAssert(junk = noErr);
			end; (* if *)
			DisposePtr(Ptr(globals));
			ICAssert(MemError() = noErr);
		end; (* if *)
		ICLMMClose := err;
	end; (* ICLMMClose *)
	
	function ICLMMOpen(globals : ComponentGlobalsPtr; self : ComponentInstance) : ComponentResult;
		var
			err: ComponentResult;
			junk : OSStatus;
	begin
		// Create our globals.
		
		globals := ComponentGlobalsPtr(NewPtrClear(sizeof(ComponentGlobals)));
		err := CheckMemError(globals);

		// Initialise our globals and tell the Component Manager about them.
		
		if err = noErr then begin
			globals^.self := self;
			globals^.target := self;
			globals^.inst := nil;
			globals^.ourResFile := 0;
			SetComponentInstanceStorage(self, Handle(globals));
		end; (* if *)

		// Clean up on failure.
				
		if err <> noErr then begin
			junk := ICLMMClose(globals, self);
			ICAssert(junk = noErr);
		end; (* if *)

		ICLMMOpen := noErr;
	end; (* ICLMMOpen *)

	function ICLMMTarget (globals: ComponentGlobalsPtr; newTarget: ComponentInstance): ComponentResult;
	begin
		globals^.target := newTarget;
		ICLMMTarget := noErr;
	end; (* ICLMMTarget *)

	(* ***** Setting Data Type ***** *)
	
	(*	These data structures are described in the Import/Export section of the code. *)
	
	type
{$PUSH}
{$ALIGN MAC68K}
		ICSetting =
			record
				magic : OSType;					// must be kICLMMCreator
				version : longint;				// kCurrentSettingVersion is currently supported
				profileID : ICProfileID;
				profileNameScript : ScriptCode;	// script code of profileName
				reserved1 : SInt16;
				profileName : Str255;
				preferenceCount : SInt32;		// count of following prefs, 0 for local settings, > 0 for external settings
				reserved2 : longint;
			end;
		ICSettingPtr = ^ICSetting;
		ICSettingHandle = ^ICSettingPtr;
		
		ICPrefHeader =
			record
				key : Str255;
				attr : ICAttr;
				dataSize : longint;
				reserved : longint;
			end;
		ICPrefHeaderPtr = ^ICPrefHeader;
		ICPrefHeaderHandle = ^ICPrefHeaderPtr;
{$ALIGN RESET}
{$POP}
	const
		kMinimumSettingSize = 8;				// sizeof(magic) + sizeof(version)
		kCurrentSettingVersion = 0;

	(* ***** Utility Routines ***** *)
	
	function ValidateSetting(setting : ICSettingHandle) : OSStatus;
		(*	This routine is called to validate that setting looks something
			like a valid ICSettingHandle.  Both local and external settings
			are considered valid as far as its concerned.
		*)
		var
			err : OSStatus;
			header : ICPrefHeader;
			prefHandleSize : longint;
			offset : longint;
			prefIndex : longint;
	begin
		
		// Basic checks for nil and purged.
		
		err := noErr;
		if (setting = nil) | (setting^ = nil) then begin
			err := paramErr;
		end; (* if *)
		
		// More basic checks, this time for "too small to extract the version number" and "bad magic".
		
		if err = noErr then begin
			prefHandleSize := GetHandleSize(Handle(setting));
			if (prefHandleSize < kMinimumSettingSize) | (setting^^.magic <> kICLMMCreator) then begin
				err := invalidSettingFakeErr;
			end; (* if *)
		end; (* if *)

		// Now do the version number conversion.  We currently only
		// have one supported version, so this is easy.
		// If you want anything else than this, you must program
		// special logic to convert settings from one version to another.
		// Of course, if you do this you must make sure that all the call
		// sites are happy with you munging the setting.
		
		if err = noErr then begin
			if setting^^.version < kCurrentSettingVersion then begin
				err := settingTooOldFakeErr;
			end; (* if *)
			if setting^^.version > kCurrentSettingVersion then begin
				err := settingTooNewFakeErr;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			if (prefHandleSize < sizeof(ICSetting)) |
					(setting^^.profileID = kICNilProfileID) |
					(setting^^.preferenceCount < 0) then begin
				err := invalidSettingFakeErr;
			end; (* if *)
		end; (* if *)
		
		// Basic checks are over.  Now check that the preference data (which is in chunks,
		// following the ICSetting header) is OK.  If preferenceCount is 0, this is
		// a local setting, so none of this code kicks in.
		
		if err = noErr then begin
			offset := sizeof(ICSetting);
			prefIndex := 1;
			while (err = noErr) & (prefIndex <= setting^^.preferenceCount) do begin
			
				// Extract the preference header for this preference
				// and do some basic consistence checks.
				
				BlockMoveData(Ptr(longint(setting^) + offset), @header, sizeof(ICPrefHeader));
				if (header.key = '') | (header.dataSize < 0) then begin
					err := invalidSettingFakeErr;
				end; (* if *)
				
				// Now advance offset based on the preference data size.
				// If we go off the end of the handle, the setting is bad.
				
				if err = noErr then begin
					offset := offset + sizeof(ICPrefHeader) + header.dataSize;
					if offset > prefHandleSize then begin
						err := invalidSettingFakeErr;
					end; (* if *)
				end; (* if *)
				prefIndex := prefIndex +1;
			end; (* if *)
			
			// Finally, make sure that offset was left at exactly
			// the size of the setting.
			
			if (err = noErr) & (offset <> prefHandleSize) then begin
				err := invalidSettingFakeErr;
			end; (* if *)
		end; (* if *)
		ValidateSetting := err;
	end; (* ValidateSetting *)
	
	function ProfileNamesEqual(const name1 : Str255; const name2 : Str255) : Boolean;
		var
			lang : LangCode;
	begin
		lang := GetScriptVariable(smSystemScript, smScriptLang);
		ProfileNamesEqual := StringOrder(name1, name2, smSystemScript, smSystemScript, lang, lang) = 0;
	end; (* ProfileNamesEqual *)

	function ResolveSettingByName(globals : ComponentGlobalsPtr;
							profileName : Str255;
							var trueProfileID : ICProfileID) : OSStatus;
		(*	This routines is called when we have a profile name and we
			want to find its corresponding profileID.  The basic algorithm
			is to go through the profiles in order comparing their names
			against profileName.  If we don't find it, trueProfileID will
			come back as kICNilProfileID /and/ we'll return icProfileNotFoundErr;
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			found : Boolean;
			profileIndex : longint;
			thisProfileID : ICProfileID;
			thisProfileName : Str255;
			{$ifc icDebug}
				perm : ICPerm;
			{$endc}
	begin
		trueProfileID := kICNilProfileID;
		err := StartUsingIC(globals);
		if err = noErr then begin
			ICAssert( (ICGetPerm(globals^.inst, perm) = noErr) & (perm = icNoPerm) );
			
			err := ICBegin(globals^.inst, icReadOnlyPerm);
			if err = noErr then begin
			
				found := false;
				profileIndex := 1;
				repeat
					err := ICGetIndProfile(globals^.inst, profileIndex, thisProfileID);
					if err = noErr then begin
						err := ICGetProfileName(globals^.inst, thisProfileID, thisProfileName);
					end; (* if *)
					if err = noErr then begin
						found := ProfileNamesEqual(profileName, thisProfileName);
						if not found then begin
							profileIndex := profileIndex + 1;
						end; (* if *)
					end; (* if *)
				until found | (err <> noErr);
				
				if found then begin
					trueProfileID := thisProfileID;
				end; (* if *)
				
				err2 := ICEnd(globals^.inst);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			
		end; (* if *)
		ResolveSettingByName := err;
	end; (* ResolveSettingByName *)
	
	function ResolveSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle;
							var trueProfileID : ICProfileID) : OSStatus;
		(*	This routines is called when we have a setting (which contains
			both a profileID and profile name and we want to find the correct
			profileID.  We needto do this because of our policy of tracking
			profiles by their profileID and, if that disappears, tracking them
			by their name.  So this routine first looks for a matching profileID
			and, if that fails, we go looking for a matching name.
			
			If we don't find it, trueProfileID will come back as kICNilProfileID
			/and/ we'll return icProfileNotFoundErr;
		*)
		var
			err : OSStatus;
			tmpName : Str255;
	begin
		trueProfileID := kICNilProfileID;
		err := StartUsingIC(globals);
		if err = noErr then begin
			err := ICGetProfileName(globals^.inst, setting^^.profileID, tmpName);
			if err = noErr then begin
				// We found it by ID, return the ID and were done.
				trueProfileID := setting^^.profileID;
			end else if err = icProfileNotFoundErr then begin
			
				// We did not find it by ID, lets try looking for it by name.
				// But we only do this if the script codes match, ie the setting
				// was made on a system with the same system script as were
				// currently running.  Otherwise we end up comparing
				// text returned by ICGetProfileName (which is, by definition,
				// in the system script), with text from another script system.
				
				if setting^^.profileNameScript = GetScriptManagerVariable(smSysScript) then begin
					err := ResolveSettingByName(globals,
												setting^^.profileName,
												trueProfileID);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		ResolveSetting := err;
	end; (* ResolveSetting *)
	
	(* ***** Import/Export ***** *)

	(*	IC LMM maintains two different types of settings.  A "local" version
		contains just the profile ID and name.  An "external" version contains
		not just that but also the values of all the Internet Config preferences.
		This is important because:
		
		a)	the user can mess with settings external to LM, so we
			just want to provide minimal pref switching, along the lines of the
			Extensions Manager LMM,

		b)	pushing around the entire IC database is expensive, in terms of
			both space (about 64K) but more importantly speed, and
			
		c)	we stil want the user to be able to export the location and have
			the entire IC database come along (so that it can be re-imported
			on some other entirely different computer).

		So, the job of converting from a local setting to a external setting,
		and vice versa, is somewhat complex.  The basic rules are:
		
		o	A local setting is just a handle that contains an ICSetting record.
			preferenceCount must be 0.

		o	An external setting is a handle that starts with an ICSetting record
			and then continues on with preferenceCount chunks, each of which
			starts with an ICPrefHeader and is followed by the data for that
			preference.  The following diagram should help:
		
			+-----------------+
			| ICSetting       | sizeof(ICSetting)
			+-----------------+
			| ICPrefHeader #1 | sizeof(ICPrefHeader)
			+-----------------+
			| data         #1 | variable
			+-----------------+
			| ICPrefHeader #2 | sizeof(ICSetting)
			+-----------------+
			| data         #2 | variable
			+-----------------+
			       ...
			+-----------------+
			| ICPrefHeader #N | sizeof(ICSetting)
			+-----------------+
			| data         #N | variable
			+-----------------+
			
			The number of chunks is recorded in the preferenceCount field of
			the ICSetting.  The size of the variable data blocks is recorded
			in dataSize field of the associated ICPrefHeader.
			
			Note that the local case is actually a degenerate version of the
			external case, ie a setting with zero embedded preferences.
	*)
	
	(*	This key is a special case that we neither import
		nor export because theyre used by IC for internal bookkeeping.
	*)
	
	const
		kICProfileName = '49434170ProfileName';
	
	(* ***** Export ***** *)
	
	(*	Export is fairly easy.  All we need to do is normalise the setting
		(ie make sure its a local setting by resizing it to sizeof(ICSetting)
		and setting preferenceCount to 0) and then loop through all off
		the IC preferences appending them to the setting.
	*)
		
	function AddPreferenceToSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle; const prefKey : Str255) : ComponentResult;
		(*	This routine is called to append a single IC preference to the
			setting.  prefKey is the key of the preference to append.  We
			assume that IC has already been switched to the correct profile.
			Also, calling this routine from inside a Begin/End pair will speed
			things up!
		*)
		var
			err : OSStatus;
			header : ICPrefHeader;
			prefData : Handle;
			attr : ICAttr;
			junkLong : longint;
	begin
		// We already store the profile name in the ICSetting, so theres
		// no point storing it again.  Plus, I dont want to confuse any
		// future version of IC by setting the name behind its back.
		// So, if were being called to store kICProfileName, just
		// return immediately.
		
		if EqualString(prefKey, kICProfileName, false, true) then begin
			err := noErr;
		end else begin
		
			// Allocate a handle to store the preference data associated with prefKey
			// and then call IC to get that data.
			
			prefData := NewHandle(0);
			err := CheckMemError(prefData);
			if err = noErr then begin
				err := ICFindPrefHandle(globals^.inst, prefKey, attr, prefData);
			end; (* if *)
			
			// Now fill out the ICPrefHeader and then call Munger to insert it
			// at the front of the prefData handle.

			if err = noErr then begin
				header.key := prefKey;
				header.attr := attr;
				header.dataSize := GetHandleSize(prefData);
				
				junkLong := Munger(prefData, 0, nil, 0, @header, sizeof(ICPrefHeader));
				err := MemError;
			end; (* if *)
			
			// Finally, HandAndHand will put prefData at the end of
			// the setting.  If that succeeds, weve added an extra
			// preference, so bump the counter.

			if err = noErr then begin
				err := HandAndHand(prefData, Handle(setting));
			end; (* if *)
			if err = noErr then begin
				setting^^.preferenceCount := setting^^.preferenceCount + 1;
			end; (* if *)

			// Clean up.
			
			if prefData <> nil then begin
				DisposeHandle(prefData);
				ICAssert(MemError = noErr);
			end; (* if *)
			
		end; (* if *)
		AddPreferenceToSetting := err;
	end; (* AddPreferenceToSetting *)

	function AddAllPreferencesToSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : OSStatus;
		(*	This routine is called to append all of the IC preferences (from the
			current IC profile) to the setting.  The caller has already made
			sure that we've switched to the right profile, and that we're inside
			an Begin/End pair.  This makes the job very easy.  We just loop
			through all the IC preferences calling AddPreferencesToSetting on
			each.
		*)
		var
			err : OSStatus;
			{$ifc icDebug}
				perm : ICPerm;
			{$endc}
			prefCount : longint;
			prefIndex : longint;
			prefKey : Str255;
	begin
		ICAssert( (ICGetPerm(globals^.inst, perm) = noErr) & (perm <> icNoPerm) );
		
		err := ICCountPref(globals^.inst, prefCount);
		prefIndex := 1;
		while (err = noErr) & (prefIndex <= prefCount) do begin
			err := ICGetIndPref(globals^.inst, prefIndex, prefKey);
			if err = noErr then begin
				err := AddPreferenceToSetting(globals, setting, prefKey);
			end; (* if *)
			if err = noErr then begin
				prefIndex := prefIndex + 1;
			end; (* if *)
		end; (* if *)
		
		AddAllPreferencesToSetting := err;
	end; (* AddAllPreferencesToSetting *)
	
	function ALMExport(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : ComponentResult;
		(*	Convert the setting from local format to external format.  This routine
			is called pretty much directly by our ALMImportExport entry point.
			The only tricky part is that we have to switch the IC current
			profile to the profile denoted by setting to make sure we export
			the right preferences (ie the ones for this profile rather than
			the ones for the current profile).
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			trueProfileID : ICProfileID;
			oldProfileID : ICProfileID;
	begin
		ICAssert(globals^.inst <> nil);
		
		// First normalise the setting by resizing it to the size of the header
		// and then resetting preferenceCount to zero.
		
		SetHandleSize(Handle(setting), sizeof(ICSetting));
		err := MemError;
		if err = noErr then begin
			setting^^.preferenceCount := 0;
		end; (* if *)
		
		// Now resolve the setting and find its correct profile ID.
		// We have to do this because theres a possibility that the
		// IC database has changed between when the setting was created
		// in ALM and when the user tries to export it.  For example,
		// you might create a location, add IC to it, delete the profile
		// in IC, and then try to export the location.
		
		if err = noErr then begin
			err := ResolveSetting(globals, setting, trueProfileID);
			if err = noErr then begin
				setting^^.profileID := trueProfileID;
			end; (* if *)
		end; (* if *)
		
		// Now append each IC preference chunk to the setting handle.  Of course,
		// IC only lets you access the current profile, so we have to switch
		// to the profile before we start getting preferences.

		if err = noErr then begin
			err := ICBegin(globals^.inst, icReadWritePerm);
		end; (* if *)
		if err = noErr then begin
			
			err := ICGetCurrentProfile(globals^.inst, oldProfileID);
			if err = noErr then begin
				err := ICSetCurrentProfile(globals^.inst, setting^^.profileID);

				if err = noErr then begin
					err := AddAllPreferencesToSetting(globals, setting);
				end; (* if *)
				
				// Switch back to the previously selected profile.
				
				err2 := ICSetCurrentProfile(globals^.inst, oldProfileID);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)

			err2 := ICEnd(globals^.inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		ALMExport := err;
	end; (* ALMExport *)

	(* ***** Import ***** *)
	
	(*	Importing a setting presents a lot more problems than exporting one.
		These include:

		1.	We have to ignore the profileID in the setting because chances 
			are that the user expects profiles to be imported/exported by name,
			not by profileID.
		
		2.	We have to ignore setting^^.profileName if setting^^.profileNameScript
			is not the system script.  If the name was created in another script
			system, theres no sense searching for it in the current IC database
			because IC makes the implicit assumption that profile names are
			in the system script.
		
		3.	We have to manage profile name conflicts, ie if youre importing
			a setting with profile name "Quinn" and a profile named "Quinn" already
			exists, what do you do?  ALM has support for a dialog giving the
			user the choice of rename, replace or cancel.  See the routine
			CheckImportedSettingName for details on how we handle this.
	
		4.	How do we merge the incoming IC settings with the existing settings.
			This is obviously an issue in the replace case, but it also crops
			up in the rename case (because IC will add a default set of
			preferences to a newly created profile).  The current policy is
			to overwrite any preferences in the destination profile that
			have preferences in the incoming setting, but to leave other
			preferences alone.
			
		5.	On a related note, what should we do about read-only preferences
			in the destination?  The current policy is to fail silently.
		
		Like I said, lots of issues.
	*)

	function CheckImportedSettingName(globals : ComponentGlobalsPtr; setting : ICSettingHandle; var createProfile : Boolean) : OSStatus;
		(*	Given the component globals and a setting to be imported, check whether
			the name in the setting collides with any current name. If it does,
			call ALMConfirmName to give the user chance to rename or replace.
			
			There are three possible return cases here:
			
			1.	Cancel -- The function returns userCanceledErr; the setting
				may have been munged in unspecified ways.
			2.	Rename -- setting^^.profileName is the new name of the setting,
				and createProfile is set to true.
			3.	Replace -- createProfile is false, setting^^.profileID is the
				ID of the profile to overwrite and setting^^.profileName is the
				new name of the profile.
			
			Note that there are non-obvious outcomes from this routine before
			of the repeat-until loop.  For example, if you import a profile "Frog"
			and it already exists in the database, you get the dialog to rename
			the profile.  If you rename the profile to "Wombat" and that already
			exists in the database, you then get the dialog again.  If you hit
			cancel this time, the function returns userCanceledErr but
			setting^^.profileName has already been changed.
		*)
		var
			err : OSStatus;
			done : Boolean;
			importChoice : integer;
			resourceInfo : integer;
			prompt : Str255;
			profileNameBackup : Str255;
	begin
		createProfile := false;
		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin

			// Have to use a local variable for prompt (instead of
			// call GetIndStr) because the lame-o-rama Location Manager
			// interfaces Im using have the prompt marked as a var
			// parameter and not a ConstStr255Param.
			
			GetIndString(prompt, rMiscStrings, strConfirmNamePrompt);

			// Check the name of the new profile against the profiles in the
			// database.  If theres no match, set createProfile and return.
			// If there is a match, ask the user what to do about it.  If
			// they hit Replace, leave createProfile false and return.  The
			// client will put the settings into setting^^.profileID.
			// If they hit rename, change setting^^.profileName and loop,
			// doing the check all over again.
			
			done := false;
			repeat
				err := ResolveSettingByName(globals, setting^^.profileName, setting^^.profileID);
				if err = icProfileNotFoundErr then begin
					err := noErr;
				end; (* if *)
				if err = noErr then begin
					if setting^^.profileID = kICNilProfileID then begin
						createProfile := true;
						done := true;
					end else begin
						profileNameBackup := setting^^.profileName;
						err := ALMConfirmName(prompt, setting^^.profileName, importChoice, nil);
						if err = noErr then begin
							case importChoice of
								ALMConfirmRenameConfig:
									(* do nothing, will cause us to loop *) ;
								ALMConfirmReplaceConfig:
									begin
										// For reasons that are not at all clear to me,
										// ALMConfirmName will set profileName to the
										// empty string if it returns ALMConfirmReplaceConfig.
										// I expected it to leave the string unchanged,
										// which would make sense if youre going to replace
										// that configuration.  The documentation has
										// nothing to say on the issue (yet!).  So I backup
										// the profileName before calling ALMConfirmName
										// and then restore the name in this case.
										// The problem is present with ALM 2.0, which
										// is my current target environment so I cant
										// just wait for this to be fixed.
										setting^^.profileName := profileNameBackup;
										done := true;
									end;
								otherwise
									begin
										ICDebugStr('Bogus importChoice!');
										err := -1;
									end;
							end; (* case *)
						end; (* if *)
					end; (* if *)
				end; (* if *)
			until (err <> noErr) | done;
		end; (* if *)
		CheckImportedSettingName := err;
	end; (* CheckImportedSettingName *)
	
	function SetPreferenceFromSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle; var offset : longint) : OSStatus;
		(*	Given a setting handle and the offset of an ICPrefHeader in that handle,
			extra the ICPrefHeader and the preference data and set that preference
			in the current profile.  We also increment offset by the number of bytes
			we extracted.  There are two special cases.  Firstly, if were being told
			to set the kICProfileName preference, we just skip it.  See the discussion
			of this in the export case.  Secondly, if we get a permission error
			while setting the preference (probably because the destination preference
			is read-only), we swallow the error.  
		*)
		var
			err : OSStatus;
			header : ICPrefHeader;
	begin
		ICAssert( btst(HGetState(Handle(setting)), kHandleLockBit) );
		ICAssert( (offset >= sizeof(ICSetting)) & (offset < GetHandleSize(Handle(setting))) );

		// Extract the header into a local variable.
				
		BlockMoveData( Ptr(longint(setting^) + offset), @header, sizeof(ICPrefHeader));

		// Some basic consistency checks.  These should all be true because
		// ValidateSetting has already checked for them.  But it pays to be sure,
		// especially when debugging nightmare code like this.
		
		ICAssert( header.key <> '' );
		ICAssert( header.dataSize >= 0 );
		
		if EqualString(header.key, kICProfileName, false, true) then begin
			err := noErr;
		end else begin
		
			// Set the preference in the IC database.
			
			err := ICSetPref(globals^.inst, header.key, header.attr, Ptr(longint(setting^) + offset + sizeof(ICPrefHeader)), header.dataSize);
			
			// Attempting to set a read-only preference will fail silently.
			
			if err = icPermErr then begin
				err := noErr;
			end; (* if *)
		end; (* if *)
		
		// Finally, adjust offset to beyond the bytes we just extracted.
		
		if err = noErr then begin
			offset := offset + sizeof(ICPrefHeader) + header.dataSize;
		end; (* if *)
		
		SetPreferenceFromSetting := err;
	end; (* SetPreferenceFromSetting *)
	
	function WriteImportedPrefsIntoProfile(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : OSStatus;
		(*	This routine is responsible for writing all of the preferences embedded
			in the setting handle into the appropriate IC profile.
			setting^^.profileID has already been resolved appropriately,
			so we can just switch to that profile, drop our data, and switch
			back.  Also note that this routien will go a lot faster if
			we're inside a Begin/End pair, but it isn't strictly necessary.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			oldProfileID : ICProfileID;
			prefIndex : integer;
			offset : longint;
	begin
		// First remember the current profile and switch to the profile
		// were supposed to be writing data to.  IC is lame in that it
		// doesnt let you write to anything but the current profile.
		// No one will see this switch because were inside an ICBegin/ICEnd
		// pair, but it will bump the seed and cause everyone to throw
		// away any cached preferences.  Such is life, kerplunk.
		
		err := ICGetCurrentProfile(globals^.inst, oldProfileID);
		if err = noErr then begin
			err := ICSetCurrentProfile(globals^.inst, setting^^.profileID);
		end; (* if *)
		
		// Now look through each chunk of preference data, setting
		// the corresponding preference in IC.
		
		if err = noErr then begin
			offset := sizeof(ICSetting);
			prefIndex := 1;
			while (err = noErr) & (prefIndex <= setting^^.preferenceCount) do begin
				err := SetPreferenceFromSetting(globals, setting, offset);
				if err = noErr then begin
					if offset > GetHandleSize(Handle(setting)) then begin
						ICDebugStr('offset out of bounds');
						err := -1;
					end; (* if *)
				end; (* if *)
				prefIndex := prefIndex + 1;
			end; (* for *)
			
			ICAssert((err <> noErr) | (offset = GetHandleSize(Handle(setting))));
			
			// Switch back to the previously selected profile.
			
			err2 := ICSetCurrentProfile(globals^.inst, oldProfileID);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		WriteImportedPrefsIntoProfile := err;
	end; (* WriteImportedPrefsIntoProfile *)
	
	function ALMImport(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : ComponentResult;
		(*	Import a setting into the IC database.  We start by trying to resolve
			which profileID we're going to use as the destination, and whether
			we need to create a new profile.  The bulk of this is handled in
			CheckImportedSettingName.  Once we have the profileID, we call
			WriteImportedPrefsIntoProfile to put the imported preferences into it.
			Finally, if all goes well we modify setting to be the local variant
			of our data structure.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			validName : Boolean;
			createProfile : Boolean;
			s : SInt8;
			haveBegun : Boolean;
	begin
		ICAssert(globals^.inst <> nil);

		s := HGetState(Handle(setting));
		HLock(Handle(setting));

		haveBegun := false;
		
		err := noErr;
		validName := ( setting^^.profileNameScript = GetScriptManagerVariable(smSysScript) );
		if validName then begin
			err := CheckImportedSettingName(globals, setting, createProfile);
		end else begin
		
			// Were going to throw away the name in the setting because its
			// in the wrong script.  Whatever new name we create will be in
			// the current system script, so we have to set profileNameScript
			// here.
			
			setting^^.profileNameScript := GetScriptManagerVariable(smSysScript);
			createProfile := true;
		end; (* if *)

		if err = noErr then begin
			err := ICBegin(globals^.inst, icReadWritePerm);
		end; (* if *)
		if err = noErr then begin
			haveBegun := true;
			if createProfile then begin
				err := ICAddProfile(globals^.inst, kICNilProfileID, setting^^.profileID);
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			if validName then begin
				err := ICSetProfileName(globals^.inst, setting^^.profileID, setting^^.profileName);
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := WriteImportedPrefsIntoProfile(globals, setting);
		end; (* if *)
		
		if haveBegun then begin
			err2 := ICEnd(globals^.inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		HSetState(Handle(setting), s);
		
		// Now mark the setting as a local variant (ie with no embedded
		// preferences) so that ALMSetCurrent doesnt have to worry
		// about the difference.
		
		if err = noErr then begin
			SetHandleSize(Handle(setting), sizeof(ICSetting));
			err := MemError;
		end; (* if *)
		if err = noErr then begin
			setting^^.preferenceCount := 0;
		end; (* if *)
		
		// Finally, on the debug version, check that weve created
		// something that we think is valid.
		
		{$ifc icDebug}
			if err = noErr then begin
				ICAssert( ValidateSetting(setting) = noErr );
			end; (* if *)
		{$endc}

		ALMImport := err;
	end; (* ALMImport *)
	
	(* ***** ALM Import/Export Entry Point ***** *)

	function ALMImportExport(globals : ComponentGlobalsPtr; import : Boolean; setting : ICSettingHandle; resRefNum : integer) : ComponentResult;
		(*	Standard ALM Entry Point
		*)
		var
			err : OSStatus;
	begin
		{$unused resRefNum}
		err := ValidateSetting(setting);
		if err = noErr then begin
			err := StartUsingIC(globals);
		end; (* if *)
		if err = noErr then begin
			if import then begin
				err := ALMImport(globals, setting);
			end else begin
				err := ALMExport(globals, setting);
			end; (* if *)
		end; (* if *)
		
		{$ifc icDebug}
			if (err <> noErr) & (err <> userCanceledErr) then begin
				ICDebugStr('ALMImportExport: Returning error (' + DecStr(err) + ') but ALM seems to silently ignore errors');
			end; (* if *)
		{$endc}
		
		ALMImportExport := err;
	end; (* ALMImportExport *)
	
	(* ***** Basic Location Manager Routines ***** *)

	function ALMGetCurrent(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : ComponentResult;
		(*	Fills out the setting handle with an ICSetting record, which represents
			the name and ID of the current profile.

			Standard ALM Entry Point
		*)
		var
			err : OSStatus;
			s : SInt8;
	begin
		err := StartUsingIC(globals);
		
		// First, resize the handle to accept our setting.
		
		if err = noErr then begin
			SetHandleSize(Handle(setting), sizeof(ICSetting));
			err := MemError;
		end; (* if *)
		
		// Then fill out the fields in the handle.
		
		if err = noErr then begin
			s := HGetState(Handle(setting));
			HLock(Handle(setting));
			
			// First set up default values, to guard against later failure.
			
			setting^^.magic := kICLMMCreator;
			setting^^.version := kCurrentSettingVersion;
			setting^^.profileID := kICNilProfileID;
			setting^^.profileName := '';
			setting^^.profileNameScript := GetScriptManagerVariable(smSysScript);
			setting^^.reserved1 := 0;
			setting^^.preferenceCount := 0;
			setting^^.reserved2 := 0;
			
			// Now get the profile ID and name from IC.
			
			err := ICGetCurrentProfile(globals^.inst, setting^^.profileID);
			if err = noErr then begin
				err := ICGetProfileName(globals^.inst, setting^^.profileID, setting^^.profileName);
			end; (* if *)

			HSetState(Handle(setting), s);
		end; (* if *)
		
		{$ifc icDebug}
			if err = noErr then begin
				ICAssert( ValidateSetting(setting) = noErr );
			end; (* if *)
		{$endc}
		
		ALMGetCurrent := err;
	end; (* ALMGetCurrent *)
	
	function ALMSetCurrent(globals : ComponentGlobalsPtr; setting : ICSettingHandle; var flags : ALMRebootFlags) : ComponentResult;
		(*	Change the IC current profile to the one in the given a setting (which
			ALM originally created with ALMGetCurrent, but it may have been saved on disk,
			imported and exported, or whatever, since then).
			
			Standard ALM Entry Point
		*)
		var
			err : OSStatus;
			trueProfileID : ICProfileID;
			currentProfileID : ICProfileID;
	begin
		flags := kALMNoChange;
		
		// Validity check setting.

		err := ValidateSetting(setting);

		// Resolve setting down to profile ID, then switch in the new
		// setting.
		
		if err = noErr then begin
			err := StartUsingIC(globals);
			
			// If we couldn't connect to IC and we're still at boot
			// time, return ALMDeferSwitchErr so that LM calls us
			// back to switch after the machine has booted.  IC is loaded
			// when components load, at the end of the INIT parade.
			// LM calls us back after the Finder is up, and we can then
			// accurately determine whether IC is installed.
			
			if (err = noICFakeErr) & (length(LMGetCurApName()^) = $FF) then begin
				err := ALMDeferSwitchErr;
			end; (* if *)
		end; (* if *)
		
		if err = noErr then begin
			err := ResolveSetting(globals, setting, trueProfileID);
		end; (* if *)		
		if err = noErr then begin
			err := ICGetCurrentProfile(globals^.inst, currentProfileID);
		end; (* if *)
		if err = noErr then begin
			if trueProfileID = currentProfileID then begin
				// Were done.  Weve already set flags to "no change",
				// so we just return noErr;
			end else begin
				err := ICSetCurrentProfile(globals^.inst, trueProfileID);
				if err = noErr then begin
					flags := kALMAvailableNow;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		ALMSetCurrent := err;
	end; (* ALMSetCurrent *)
	
	function ALMCompareSetting(globals : ComponentGlobalsPtr; 
								setting1 : ICSettingHandle; 
								setting2 : ICSettingHandle; 
								var equal : Boolean) : ComponentResult;
		(*	Standard ALM Entry Point
		*)
		var
			err : OSStatus;
			name1 : Str255;
			name2 : Str255;
	begin
		{$unused globals}
		err := ValidateSetting(setting1);
		if err = noErr then begin
			err := ValidateSetting(setting1);
		end; (* if *)
		if err = noErr then begin
			name1 := setting1^^.profileName;
			name2 := setting2^^.profileName;
			equal := (setting1^^.profileID = setting2^^.profileID) &
						(setting1^^.profileNameScript = setting2^^.profileNameScript) &
						ProfileNamesEqual(name1, name2);
		end; (* if *)
		ALMCompareSetting := err;
	end; (* ALMCompareSetting *)
	
	function ALMEditSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : ComponentResult;
		(*	Our response to being asked to edit the setting is to simple
			call the ICEditPreferences API, which will launch IC and open
			the corresponding preferences file.

			Standard ALM Entry Point
		*)
		var
			err : OSStatus;
	begin
		{$unused setting}
		
		err := StartUsingIC(globals);
		if err = noErr then begin
			err := ICEditPreferences(globals^.inst, '');
			
			// Map error codes to their more specific fake error codes,
			// so that our describe error routine knows which specific
			// error to describe.
			
			case err of
				memFullErr:
					err := memFullFakeErr;
				afpItemNotFound:
					err := afpItemNotFoundFakeErr;
				noPortErr:
					err := noPortFakeErr;
				otherwise	
					(* do nothing *) ;
			end; (* case *)
		end; (* if *)

		// Note that IC will not bring Location Manager back to front when editing
		// is done, which the documentation says it should do.  However, this is
		// too tricky for me to organise at this stage.
		
		ALMEditSetting := err;
	end; (* ALMEditSetting *)
	
	function ALMDescribeSettings(globals : ComponentGlobalsPtr; 
									setting : ICSettingHandle; 
									text : CharsHandle) : ComponentResult;
		(*	Standard ALM Entry Point
		*)
		var
			err : OSStatus;
			resourceInfo : integer;
			descriptionString : Str255;
	begin
		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			err := ValidateSetting(setting);
			if err = noErr then begin
				descriptionString := concat(
										GetIndStr(rMiscStrings, strInternetConfigSet),
										setting^^.profileName);
				SetHandleSize(Handle(text), length(descriptionString));
				err := MemError;
			end; (* if *)
			if err = noErr then begin
				BlockMoveData(@descriptionString[1], text^, length(descriptionString));
			end; (* if *)
		end; (* if *)

		// Clean up.
				
		EndUsingResources(resourceInfo);
		
		ALMDescribeSettings := err;
	end; (* ALMDescribeSettings *)
	
	function ALMDescribeError(globals : ComponentGlobalsPtr; lastErr : OSErr; var errStr : Str255) : ComponentResult;
		(*	Standard ALM Entry Point
		*)
		var
			err : OSStatus;
			stringIndex : integer;
			resourceInfo : integer;
	begin
		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			case OSStatus(lastErr) of
				noICFakeErr :
					stringIndex := strNoICErr;
				insufficientICFakeErr :
					stringIndex := strInsufficientICErr;
				memFullFakeErr : 
					stringIndex := strNoMemoryErr;
				afpItemNotFoundFakeErr :
					stringIndex := strCantFindICErr;
				noPortFakeErr :
					stringIndex := strCantHackIt;
				icProfileNotFoundErr :
					stringIndex := strProfileNotFoundErr;
				invalidSettingFakeErr :
					stringIndex := strInvalidSetting;
				settingTooOldFakeErr :
					stringIndex := strSettingTooOld;
				settingTooNewFakeErr :
					stringIndex := strSettingTooNew;
				otherwise
					err := paramErr;		// Ask Location Manager to come up with an error.
			end; (* case *)
		end; (* if *)
		if err = noErr then begin
			errStr := GetIndStr(rMiscStrings, stringIndex);
		end; (* if *)
		
		EndUsingResources(resourceInfo);

		ALMDescribeError := err;
	end; (* ALMDescribeError *)

	function ALMGetScriptInfo(globals : ComponentGlobalsPtr; var info : ALMScriptMgrInfo) : ComponentResult;
		(*	Note that we don't use the standard 'intl' resource
			because its kinda lame.  Specifically, there's no
			point have a version number in the resource because the
			resource cant be revised without revising the code.
			Secondly, the font name field was too small.  So instead
			I defined by own resource type 'SINF' and a Pascal
			type (ScriptInfo) to match it.
			
			Standard ALM Entry Point
		*)
{$PUSH}
{$ALIGN MAC68K}
		type
			ScriptInfo =
				record
					script   : ScriptCode;
					region   : SInt16;
					language : LangCode;
					fontSize : SInt16;
					fontName : Str255;
				end;
			ScriptInfoPtr = ^ScriptInfo;
			ScriptInfoHandle = ^ScriptInfoPtr;
{$ALIGN RESET}
{$POP}
		var
			err : OSStatus;
			resourceInfo : integer;
			scriptInfoRsrc : ScriptInfoHandle;
			fontName : Str255;
	begin
		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			scriptInfoRsrc := ScriptInfoHandle(Get1Resource('SINF', rOurScriptInfo));
			err := CheckResError(scriptInfoRsrc);
		end; (* if *)
		if err = noErr then begin
			fontName := scriptInfoRsrc^^.fontName;
			
			info.version := kALMScriptInfoVersion;
			info.scriptCode := scriptInfoRsrc^^.script;
			info.regionCode := scriptInfoRsrc^^.region;
			info.langCode   := scriptInfoRsrc^^.language;
			GetFNum(fontName, info.fontNum);
			info.fontSize   := scriptInfoRsrc^^.fontSize;
		end; (* if *)
		
		// Clean up.
				
		EndUsingResources(resourceInfo);

		ALMGetScriptInfo := err;
	end; (* ALMGetScriptInfo *)
	
	function ALMGetInfo(globals : ComponentGlobalsPtr; var text : CharsHandle; var style : StScrpHandle; filter : ModalFilterUPP) : ComponentResult;
		(*	Standard ALM Entry Point
		*)
		var
			err : OSStatus;
			tmp : Handle;
			resourceInfo : integer;
	begin
		{$unused filter}

		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			tmp := Get1Resource('TEXT', rInfoText);
			err := CheckResError(tmp);
			if err = noErr then begin
				DetachResource(tmp);
				ICAssert(ResError = noErr);
				text := CharsHandle(tmp);
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			tmp := Get1Resource('styl', rInfoText);
			err := CheckResError(tmp);
			if err = noErr then begin
				DetachResource(tmp);
				ICAssert(ResError = noErr);
				style := StScrpHandle(tmp);
			end; (* if *)
		end; (* if *)
		
		// Clean up.
		
		EndUsingResources(resourceInfo);

		ALMGetInfo := err;
	end; (* ALMGetInfo *)

	(* ***** Component Entry Point ***** *)

	function Main(var params : ComponentParameters; storage : Ptr) : ComponentResult;
		(* The main component entry point.  Its just a standard component dispatcher. *)
		var
			oldA4 : longint;
			oldA5 : longint;
			proc: ProcPtr;
			result : ComponentResult;
	begin
		(*	Bring in the correct A4 world.  Also, set up A5 to point to the current
			applications A5 world.  We dont use A5 based globals, except insofar as
			system API routines use them.  So we dont need to create our own A5 world.  
			But we do need to make sure our A5 world is correct, because we may call 
			system routines that need A5 to point to some valid A5 world.
		*)
		oldA4 := SetCurrentA4;
		oldA5 := SetCurrentA5;

		(* Setup proc based on which routine is to be called. *)
		
		proc := nil;
		case params.what of
			// Component Manager stuff
			
			kComponentVersionSelect: 
				result := 0;			// 
			kComponentCanDoSelect: 
				proc := @ICLMMCanDo;
			kComponentOpenSelect: 
				proc := @ICLMMOpen;
			kComponentCloseSelect: 
				proc := @ICLMMClose;
			kComponentTargetSelect: 
				proc := @ICLMMTarget;
				
			// Stuff specific to Location Manager Modules

			kALMGetCurrentSelect:
				proc := @ALMGetCurrent;
			kALMSetCurrentSelect:
				proc := @ALMSetCurrent;
			kALMCompareSettingSelect:
				proc := @ALMCompareSetting;
			kALMEditSettingSelect:
				proc := @ALMEditSetting;
			kALMDescribeSettingsSelect:
				proc := @ALMDescribeSettings;
			kALMDescribeErrorSelect:
				proc := @ALMDescribeError;
			kALMImportExportSelect:
				proc := @ALMImportExport;
			kALMGetScriptInfoSelect:
				proc := @ALMGetScriptInfo;
			kALMGetInfoSelect:
				proc := @ALMGetInfo;
			
			otherwise
				result := badComponentSelector;
		end; (* case *)

		(* If weve got a routine to call, call it.  Note that the default case
			above has set result up to return the appropriate error code if
			proc is nil.
		*)
		if proc <> nil then begin
			result := CallComponentFunctionWithStorage(Handle(storage), params, proc);
		end; (* if *)

		(* Clean up A5 and A4. *)

		oldA5 := SetA5(oldA5);
		oldA4 := SetA4(oldA4);
		Main := result;
	end; (* Main *)

end. (* ICComponent *)